#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <time.h>

// CSR lib
#include "csrOpt.h"
#include "csrTypes.h"
#include "csrMath.h"
#include "csrMem.h"
#include "csrIO.h"
#include "csrComb.h"

// Valve formats
#include "vdf.h"
#include "vpk.h"
#include "vfilesys.h"

#include "vmdl.h"
#include "vmf.h"

// CSR main
#include "csr32f.h"
#include "csrDraw.h"

#define CSR_VERSION "0.0.1"

// gcc -Wall -fsanitize=address csRadar.c -o csRadar -lm

int main( int argc, char *argv[] )
{
	char *arg;
	char *strings[ 20 ];
	int num_strings = 0;
	
	float padding = 128.f;
	u32 resolution = 1024;
	int standard_layers = 0;
	int write_txt = 1;
	int multi_sample = 1;
	char output_path[ 512 ];		// Full path eg. /home/harry/my_map.vmf
	char vmf_name[ 128 ];			// Just the base name eg. my_map
	int output_set = 0;

	while( csr_argp( argc, argv ) )
	{
		if( (arg = csr_arg()) )
		{
			if( num_strings == 20 )
			{
				fprintf( stderr, "Too many arguments! Max 20\n" );
				fs_exit();
				exit(0);
			}
			
			printf( "got: %s\n", arg );
			
			strings[ num_strings ++ ] = arg;
		}
		
		if( (arg = csr_opt_arg( 'o' )) )
		{
			strcpy( output_path, arg );
			csr_path_winunix( output_path );
			
			output_set = 1;
		}
		
		if( (arg = csr_opt_arg( 'g' )) )
		{
			fs_set_gameinfo( arg );
		}
		
		if( (arg = csr_opt_arg( 'r' )) )
		{
			resolution = atoi( arg );
		}
		
		if( (arg = csr_long_opt_arg( "padding" )) )
		{
			padding = atof( arg );
		}

		if( csr_opt( 'v' ) || csr_long_opt( "version" ) )
		{
			printf( "csRadar version: " CSR_VERSION "\n" );
			return 0;
		}
		
		if( csr_opt( 'h' ) || csr_long_opt( "help" ) )
		{
			// Display help
			printf
			( 
				"csRadar Copyright (C) 2021 Harry Godden (hgn)\n"
				"\n"
				"Usage: ./csRadar map.vmf -g \"/gamedir/gameinfo.txt\" layout cover\n"
				"   VMF file is first, then any other arguments (eg. layout, cover), will specify\n"
				"   visgroups to be rendered into individual files\n"
				"   No visgroups specified will simply draw everything\n"
				"\n"
				"Options:\n"
				"   -g <gameinfo.txt path>         Required if you are loading models\n"
				"   -r 1024                        Output resolution\n"
				"   -o <output>                    Specify output name/path\n"
				"   --padding=128                  When cropping radar, add padding units to border\n"
				"   --standard-layers              Use standard TAR layers/groups\n"
				"   --no-txt                       Don't create matching radar txt\n"
				"   --multi-sample=                [ none, 2, 4, 4r, 8kn (default), 16c ]\n"
				"\n"
				"   -v --version                   Display program version\n"
				"   -h --help                      Display this help text\n"
			);
			
			return 0;
		}
	}
	
	if( num_strings )
	{
		vmf_map *map = vmf_init( strings[0], 1 );
	
		if( map )
		{
			// Path handling
			if( !output_set )
			{
				strcpy( output_path, strings[0] );
				csr_stripext( output_path );
			}
			
			char *base_name;
			if( !(base_name = csr_findext( output_path, '/' ) ))
			{
				base_name = output_path;
			}
			
			strcpy( vmf_name, base_name );
			
			printf( "output_path: '%s'\nvmf_name: '%s'\n", output_path, vmf_name );
			
			
			// Main
					
			csr_target target;
			
			csr_create_target( &target, resolution, resolution );
			csr_rt_clear( &target );
			
			// Compute bounds
			csr_filter filter = 
			{
				.classname = NULL,
				.visgroup = NULL,
				.compute_bounds_only = 1
			};
			
			// One pass for fitting, second pass for drawing
			for( int i = 0; i < 2; i ++ )
			{
				if( num_strings == 1 )
				{
					// Draw everything
					draw_vmf_group( &target, map, map->root, &filter, NULL, NULL );
					csr_rt_save_buffers( &target, output_path, "all" );
				}
				else
				{
					// Draw groups
					for( int i = 1; i < num_strings; i ++ )
					{
						filter.visgroup = strings[ i ];
						
						draw_vmf_group( &target, map, map->root, &filter, NULL, NULL );
						csr_rt_save_buffers( &target, output_path, strings[i] );
						
						csr_rt_clear( &target );
					}
				}
				
				filter.compute_bounds_only = 0;
				csr_auto_fit( &target, padding );
			}
			
			if( write_txt )
			{
				char txt_path[ 512 ];
				
				strcpy( txt_path, output_path );
				strcat( txt_path, ".txt" );
				
				csr_write_txt( txt_path, vmf_name, &target );
			}
			
			csr_rt_free( &target );
			vmf_free( map );
		}
		else
		{
			fprintf( stderr, "Could not load VMF\n" );
		}
	}
	else
	{
		fprintf( stderr, "Missing required argument: mapfile\n" );
	}
	
	fs_exit();

	return 0;
}
